print("import: 'pygments'")
import pygments

print("import: 'pygments.filters'")
import pygments.filters

print("import: 'pygments.formatters'")
import pygments.formatters

print("import: 'pygments.lexers'")
import pygments.lexers

print("import: 'pygments.styles'")
import pygments.styles

print("import: 'pygments.plugin'")
import pygments.plugin

